/***
*		keyolator.c -- tool for b&p to translate notes to a new key
*		 1994, 1996  Another Option
***/

/*!! processnote() still not verified !!*/

/*
============================================================================
|    DATE   | WHAT
============================================================================
   2 Apr 94 | created from lockinkey
  13 Apr 94 | added source key selection
  24 Apr 94 | fixed (?) conversion to/from scale
  24 Feb 95 | recompile for new bozo request position under pointer.
  16 Nov 96 | stdargs/regargs + poppipes for routing 
*/

#define DBUG 0

/* define this to write to sd0:BP.dump instead of con: for dprintf */
//#define FILEDUMP 1

#include <exec/exec.h>
#include <dos/dos.h>
#include <intuition/intuition.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <string.h>
#include <clib/alib_protos.h>
#include <clib/alib_stdio_protos.h>
//#include <stdio.h>

#include "bars.h"
#include "edit.h"
#include "bplib.h"
#include "barsmacros.h"
#include "keyolatorw.h"
#include "proto/keyolator_protos.h"
#include "/icons/KeyolatorDown.h"
#include "/icons/KeyolatorUp.h"
#include "snapshotgadget.h"
#include "popuppipes.h"
#include "toolid.h"
#include "debug.h"

static UBYTE const versiontag[] = "$VER: Keyolator 40.1 (18.11.96)";

/* size of the noteon/off lookup table */
#define TABLE_SIZE 128

struct KeyTool {
	struct Tool tool;
	struct Chord *newkey; /* pointer to the destination key */
	struct Chord *oldkey; /* the source key */
	UBYTE  newroot;       /* tonic of the scale */
	BYTE   newflats;
	UBYTE  oldroot;
	BYTE   oldflats;
	UBYTE  flags;
	UBYTE  moflags;
	UWORD  _word_pad;
	char   noteout[TABLE_SIZE]; /* map of what notes got output */
};

/*** flag bits ***/
/* get new/old scales from KeyEvents */
#define NEW_KEY_EVENTS 1
#define OLD_KEY_EVENTS 2
//!! branching not implemented !!
#define USE_SYSEX      4
#define FREE_SYSEX     8
/* if set, inittool() has to resolve tool->key after a load from disk */
#define FLAG_BOGUS  0x80

/* mo' flags */
#define BRANCH_OLD     1
#define PASS_OLD       2
#define BRANCH_NEW     4
#define PASS_NEW       8

#define SX_LENGTH 14

#define DEFAULT_FLAGS (NEW_KEY_EVENTS | PASS_NEW)

#define TOOL_SIZE sizeof(struct KeyTool)

extern struct Functions *functions;

/*** bozo request stuff ***/
#define BOZO_GTEXT "Continue"

/***
void dumptool(struct KeyTool *tool)
{
	dprintf("\ntool @ 0x%8X  flags=%s\n",tool,showbits(tool->flags,8));
	dprintf(" newkey=0x%8X newroot=%2d newflats=%d ",
	         tool->newkey,tool->newroot,tool->newflats);
	if (tool->newkey) {
		dprintf(" (%s) %s",functions->stringtext(tool->newkey->name),showbits(tool->newkey->pattern,12));
	}
	else dprintf(" newkey not selected");
	dprintf("\n");
	dprintf(" oldkey=0x%8X oldroot=%2d oldflats=%d ",
	         tool->oldkey,tool->oldroot,tool->oldflats);
	if (tool->oldkey) {
		dprintf(" (%s) %s",functions->stringtext(tool->oldkey->name),showbits(tool->oldkey->pattern,12));
	}
	else dprintf(" oldkey not selected");
	dprintf("\n");
}
***/

/*==========================================================================
* indextokey
*
*  INPUT: long
* RETURN: pointer to Chord in scalelist, or NULL if not found
*
* EFFECT: scans and counts
*=========================================================================*/

static struct Chord *__regargs indextokey(long i)

{
	register struct Chord *c;

	if (i > -1) {
		for ( c = functions->scalelist; c ; c = c->next ) {
			if (!(--i)) {
//dprintf("itok()=%ld i=%ld\n",c,i);
				return(c);
			}
		}
	}
	return(NULL); // key removed or not selected
}

/*==========================================================================
* inittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: sets tool defaults
*=========================================================================*/

static void __regargs inittool(struct KeyTool *tool)

{
	if (tool->flags & FLAG_BOGUS) {
		tool->flags &= ~FLAG_BOGUS;
		if (((long)(tool->newkey)) != -1) { /* then this is the index of the key */
			tool->newkey = indextokey((long)tool->newkey);
			if (!tool->newkey) {
#pragma msg 104 ignore push
				BozoRequest(NULL,WTITLE,
		       "Cannot find the new key that was selected\nwhen the tool was saved!",
		       0,-1,-1,0,BOZO_GTEXT,NULL);
#pragma msg 104 ignore pop
			}
		}
		else tool->newkey = NULL; /* not selected when saved */
		if (((long)(tool->oldkey)) != -1) {
			tool->oldkey = indextokey((long)tool->oldkey);
			if (!tool->oldkey) {
#pragma msg 104 ignore push
				BozoRequest(NULL,WTITLE,
		       "Cannot find the old key that was selected\nwhen the tool was saved!",
		       0,-1,-1,0,BOZO_GTEXT,NULL);
#pragma msg 104 ignore pop
			}
		}
		else tool->oldkey = NULL; /* not selected when saved */
	}
	else {
		tool->newkey = tool->oldkey = NULL;
		tool->newroot = tool->oldroot = 0;
		tool->newflats = tool->oldflats = 0;
		tool->flags = DEFAULT_FLAGS;
	}
	/* clear the lookup table */
	memset((void *)(&tool->noteout[0]),-1,128);
	tool->tool.touched |= TOUCH_INIT;
}

/*==========================================================================
* scaletotwelve
*
*  INPUT: pointer to Chord, scale root, flats flag, note value
* RETURN: midi note number
*
* EFFECT: works like b&p scaletotwelve, but gets different arguments (the
*         real poop instead of a pointer to a Clip) so it doesn't get a key
*         from timetokey()
*=========================================================================*/

static long __regargs scaletotwelve(struct Chord *key, BYTE root, BYTE value)

{
	unsigned long pattern;
	ULONG bit;
	char i;
	char base = 0;
	char octsize;

//dprintf("\nscaletotwelve() root=%d pattern=",root);
//dprintf("scaletotwelve() value=%d\n",value);
	if (key) {
		pattern = key->pattern;
		octsize = key->count;
	}
	else { /* use Cmaj */
		root = 0;
		pattern = 0xAB5;
		octsize = 7;
	}
	pattern |= (pattern << 12);
	pattern >>= (12-root);
	pattern &= (0xFFFF << root);
//dprintf("%s ",showbits(pattern,16));
	for ( ; value >= octsize ; ) {
		value -= octsize;
		base += 12;
	}
//dprintf("val=%d base=%d\n",value,base);
	for ( i=root ; i < 16+root ; i++ ) {
		bit = 1<<i;
		if (pattern & bit) {
//dprintf("bit %ld hit value=%d\n",i,value);
			if (!value) {
//dprintf("sto12() return1 i=%d base=%d\n\n",i,base);
				return((long)(i + base));
			}
			value--;
		}
	}
//dprintf("sto12() return2 base=%d\n",base);
	return((long)base + root);
}

/*==========================================================================
* twelvetoscale
*
*  INPUT: pointer to Chord, scale root, flats flag, value, pointer to accidental
* RETURN: char
*
* EFFECT: b&p twelvetoscale(), but gets scale/key poop instead of pointer to
*         clip; doesn't call timetokey()
*=========================================================================*/

static long __regargs twelvetoscale(struct Chord *key, BYTE root, BYTE flats, 
                                    BYTE value, BYTE *offset)
{
	unsigned long pattern;
	char i;
	char j = -1;
	char base;
	char octsize;
	char sharp = 0;
	ULONG bit;

//dprintf("twelvetoscale() in_value=%d root=%d pattern=",value,root);
//dprintf("twelvetoscale() in_value=%d\n",value);
	if (key) {
		pattern = key->pattern;
		octsize = key->count;
	}
	else {
		root = 0;
		pattern = 0xAB5;
		octsize = 7;
		flats = 0;
	}
	if (!pattern) return(0);
	pattern |= (pattern << 12);
	pattern |= (pattern << 24);
	pattern = pattern >> (12-root);
	pattern = pattern & (0xFFF << root);
//dprintf("%s ",showbits(pattern,12+root));

	value -= root;
	base = 0;
	for ( ; value > 11 ; ) {
		value -= 12;
		base += octsize;
	}
//dprintf("val=%d base=%d\n",value,base);

	for ( i=root ; i <= 24 ; i++ ) {
		bit = 1 << i;
		if (pattern & bit) {
			j++;
			sharp = 0;
//dprintf(" bit %d hit j=%d value=%d\n",i,j,value);
		}
		else {
			sharp++;
		}
		if (!value) break;
		value--;
	}
//dprintf(" loop exit  i=%d value=%d\n",i,value);
	if (flats) {
		if (sharp) {
			sharp = 0;
			j++;
			for ( ; i<24 ; i++) {
				bit = 1 << i;
				if (pattern & bit) break;
				sharp--;
			}
		}
		if (sharp < -12) sharp += 12;
	}
//dprintf("12tos() return j=%d base=%d sharp=%d\n\n",j,base,sharp);
	if (offset) *offset = sharp;
	return((long)j + base);
}

/*==========================================================================
* changenote
*
*  INPUT: necessary stuff
* RETURN: none
*
* EFFECT: input notenum gets changed to a new key
*=========================================================================*/

static UBYTE __regargs changenote(struct Chord *newkey, UBYTE newroot,
                                  BYTE newflats, struct Chord *oldkey,
											 UBYTE oldroot, BYTE oldflats, UBYTE notenum)
{
	register long value;
//	long rootvalue = 0;
	BYTE accidental;
//	BYTE accidental2 = 0;
	static struct Chord major = {
		NULL,NULL,0xAB5,0,7
	};

	if (!oldkey) {
//dprintf("--no oldkey\n");
		oldkey = &major;
		oldroot = 0;
		oldflats = 0;
	}
	if (!newkey) {
//dprintf("--no newkey\n");
		newkey = &major;
		newroot = 0;
		newflats = 0;
	}
/*
dprintf("cn() old key=%8X root=%2d flats=%1d    note=%3d\n",
                                         oldkey,oldroot,oldflats,notenum);
dprintf("     new key=%8X root=%2d flats=%1d\n",newkey,newroot,newflats);
*/
	value = twelvetoscale(oldkey,oldroot,oldflats,notenum,&accidental);

//dprintf("val=%ld rootval=%ld acc=%d acc2=%d",value,rootvalue,accidental,accidental2);
	value = scaletotwelve(newkey,newroot,value);
	value += accidental;
	while (value < 0) value += 12;
	while (value > 127) value -= 12;
//dprintf("    return=%3ld\n\n",value);
	return((UBYTE)value);
}

/*==========================================================================
* findkey
*
*  INPUT: address of scale to try to locate
* RETURN: pointer to Chord
*
* EFFECT: scans functions->scalelist for a matching address
*=========================================================================*/

static struct Chord * __regargs findkey(struct Chord *scale)

{
	register struct Chord *k;

	for ( k = functions->scalelist ; k ; k=k->next ) {
		if (k == scale) {
			return(k);
		}
	}
	/* else */
	return(NULL);
}
	
/*==========================================================================
* processlist
*
*  INPUT: pointers to tool and list to process, start and stop times
* RETURN: pointer to inverted list
*
* EFFECT: alters note numbers in list to conform with the selected key/scale.
*         If the tool is set to respond to KeyEvents, it will start with
*         the current key/scale, which will be used until it encounters a
*         KeyEvent in the list.  If the tool is set to use a particular user
*         selected key/scale, it will refuse to process the list if that key
*         is no longer present in functions->scalelist.
*=========================================================================*/

static struct Event * __saveds __stdargs processlist(struct KeyTool *tool,
                                           struct Event *list,
                                           long start, long stop)
{
	register struct Event *e = list;
	struct KeyEvent *key;
	struct Chord *newscale = NULL;
	struct Chord *oldscale = NULL;
	struct Event *copy;
	struct Event *copylist = NULL;
	UBYTE newroot = 0;
	BYTE newflats = 0;
	UBYTE oldroot = 0;
	BYTE oldflats = 0;
	struct Clip *clip;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);
	if (start < 0) start=0;
	while (e->time < start) e=e->next;

/******* this is damned near duped for processevent()! ********/

	if (tool->tool.track->edit && tool->tool.track->edit->inedit) {
		clip = &tool->tool.track->edit->clip;
	}
	else {
		clip = tool->tool.clip;
	}
	if (tool->flags & NEW_KEY_EVENTS) {
		key = functions->timetokey(clip,start);
		if (key) {
			newscale = key->scale;
			newroot = key->root;
			newflats = key->flats;
		}
	}
	else if (tool->newkey) {
		if (newscale = findkey(tool->newkey)) {
			newroot = tool->newroot;
			newflats = tool->newflats;
		}
		else { /* scale not in list */
			functions->msgbox("The selected new key was removed!",
			                  "Hmmm",NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
			return(list);
		}
	}
	if (tool->flags & OLD_KEY_EVENTS) {
		key = functions->timetokey(clip,start);
		if (key) {
			oldscale = key->scale;
			oldroot = key->root;
			oldflats = key->flats;
		}
	}
	else if (tool->oldkey) {
		if (oldscale = findkey(tool->oldkey)) {
			oldroot = tool->oldroot;
			oldflats = tool->oldflats;
		}
		else { /* scale not in list */
			functions->msgbox("The selected old key was removed!",
			                  "Hmmm",NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
			return(list);
		}
	}
/*
if (oldscale && oldscale->name) dprintf("old=%X (%s)",oldscale,functions->stringtext(oldscale->name));
dprintf("\n");
*/

/**************************/

	for ( ; e && e->time <= stop ; e=e->next ) {
		if (IsNoteOn(e) || IsPolyTouch(e)) {
			if ((tool->moflags & PASS_OLD) || 
			    ((tool->moflags & BRANCH_OLD) && tool->tool.branch)) {
				copy = functions->dupeevent(e);
				if (copy) {
					if (tool->moflags & PASS_OLD) copy->tool = tool->tool.next;
					else {
						copy->tool = tool->tool.branch;
						copy->type |= EVENT_BRANCH;
					}
					copy->next = copylist;
					copylist = copy;
				}
			}
			if (tool->flags & NEW_KEY_EVENTS) {
				key = functions->timetokey(clip,e->time);
				if (key) {
					newscale = key->scale;
					newroot = key->root;
					newflats = key->flats;
				}
				else newscale = NULL;
			}
			else if (tool->flags & OLD_KEY_EVENTS) {
				key = functions->timetokey(clip,e->time);
				if (key) {
					oldscale = key->scale;
					oldroot = key->root;
					oldflats = key->flats;
				}
				else oldscale = NULL;
			}
/*
if (newscale && newscale->name) dprintf("new=%X root=%2d (%s)",
             newscale,newroot,functions->stringtext(newscale->name));
dprintf("\n");
if (oldscale && oldscale->name) dprintf("old=%X root=%2d (%s)",
             oldscale,oldroot,functions->stringtext(oldscale->name));
dprintf("\n");
*/
			e->byte1 = changenote(newscale,newroot,newflats,
			                      oldscale,oldroot,oldflats,e->byte1);
			if ((tool->moflags & BRANCH_NEW) && tool->tool.branch) {
				e->tool = tool->tool.branch;
				e->type |= EVENT_BRANCH;
			}
			else {
				e->tool = tool->tool.next;
			}
		}
	}
	if (copylist) {
		list = (struct Event *)functions->List_Cat((struct ListItem *)list,
		                                           (struct ListItem *)copylist);
		list = functions->sorteventlist(list);
	}
	return(list);
}

#define tw tool->tool.window

#define SECRET_MESSAGE 9999 // mn_Length of message to edittool() for refresh

/*==========================================================================
* dosysex
*
*  INPUT: pointer to StringEvent
* RETURN: pointer to Event
*
* EFFECT: gets new tool state from sysex message
*=========================================================================*/

static __inline struct Event *dosysex(struct KeyTool *tool,
                                      struct StringEvent *event)

{
	register unsigned char *sxdata;
	struct Message *msg; // send a message to edittool() to refresh gadgets' states
	UWORD sxflag;        // preserve bits for using/freeing sysex messages

	if (tool->flags & USE_SYSEX) {
		if (sxdata = checksysex(event,TOOL_SYSEX_ID,tool->tool.toolid)) {
			sxflag = tool->flags & (USE_SYSEX | FREE_SYSEX);
			memcpy((void *)&tool->newkey,(void *)sxdata,SX_LENGTH);
			tool->newkey = indextokey((long)tool->newkey);
			tool->oldkey = indextokey((long)tool->oldkey);
			if (sxflag & FREE_SYSEX) tool->flags |= FREE_SYSEX;
			if (tool->tool.window) {
				/* window is open; send a secret message to refresh */
				if (msg=(struct Message *)functions->myalloc(sizeof(struct Message),0)) {
					msg->mn_Node.ln_Type = NT_MESSAGE;
					msg->mn_ReplyPort = NULL;
					msg->mn_Length = SECRET_MESSAGE;
					PutMsg(tool->tool.window->UserPort,msg);
				}
			}
			if (tool->flags & FREE_SYSEX) {
				functions->freestring(event->string);
				functions->freeevent((struct Event *)event);
				event=NULL;
			}
		}
	}
	return((struct Event *)event);
}

/*==========================================================================
* processevent
*
*  INPUT: pointer to Event
* RETURN: pointer to Event
*
* EFFECT: translates notes from one key to a different key.  frees all noteoffs 
*         to avoid recalculation and in case tool state changes between on and 
*         off; sticks the resulting note into tool->noteout[].  resets tool->~key
*         fields if selected key comes up missing.
*         SHOULD REFRESH AN OPEN TOOL WINDOW IF THE KEY IS MISSING.
*=========================================================================*/

static struct Event * __saveds __stdargs processevent(struct Event *event)

{
   struct KeyTool *tool = (struct KeyTool *)event->tool;
	struct Chord *newscale = NULL;
	struct Chord *oldscale = NULL;
	struct Clip *clip;
	struct KeyEvent *key;
	char notein;
	UBYTE newroot = 0;
	UBYTE oldroot = 0;
	BYTE newflats = 0;
	BYTE oldflats = 0;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);

	if (IsAllOff(event)) { /*?? need to send offs for no response to alloff ??*/
//dprintf("ALLOFF: @ %6ld\n",event->time);
		memset((char *)tool->noteout,-1,TABLE_SIZE); /* reset table entries */
		event->tool = tool->tool.next;
	}
	else if (IsNoteOff(event)) { // doesn't copy noteoffs yet!!!!!!!!!!!!!!!!!

//dprintf("OFF: @ %6ld in=%3d out=%3d\n",event->time,event->byte1,
//                            tool->noteout[event->byte1]);

		if (tool->noteout[event->byte1] > -1) {
			event->byte1 = tool->noteout[event->byte1];
//			tool->noteout[event->byte1] = -1;
		}
		event->tool = tool->tool.next;
	}
	else if (IsNoteOn(event) || IsPolyTouch(event)) {
/*!! noteons should be cloned; noteoffs should be freed in case of a change !!*/
		if ((tool->flags & NEW_KEY_EVENTS) || (tool->flags & OLD_KEY_EVENTS)) {
			if (tool->tool.track->edit && tool->tool.track->edit->inedit) {
				clip = &tool->tool.track->edit->clip;
			}
			else {
				clip = &tool->tool.track->clip;
			}
		}

		if (tool->flags & NEW_KEY_EVENTS) {
			if (key = functions->timetokey(clip,event->time)) {
				newscale = key->scale;
				newroot = key->root;
				newflats = key->flats;
			}
		}
		else if (tool->newkey) {
			if (newscale = findkey(tool->newkey)) {
				newroot = tool->newroot;
				newflats = tool->newflats;
			}
			else {
				newscale = tool->newkey = NULL;
			}
		}
		if (tool->flags & OLD_KEY_EVENTS) {
			if (key = functions->timetokey(clip,event->time)) {
				oldscale = key->scale;
				oldroot = key->root;
				oldflats = key->flats;
			}
		}
		else if (tool->oldkey) {
			if (oldscale = findkey(tool->newkey)) {
				oldroot = tool->newroot;
				oldflats = tool->newflats;
			}
			else {
				oldscale = tool->oldkey = NULL;
			}
		}
		if ((tool->moflags & PASS_OLD) || 
		    ((tool->moflags & BRANCH_OLD) && tool->tool.branch)) {
			event->next = functions->dupeevent(event);
			if (event->next) {
				if (tool->moflags & PASS_OLD) event->next->tool = tool->tool.next;
				else {
					event->next->tool = tool->tool.branch;
					event->next->type |= EVENT_BRANCH;
				}
				event->next->next = NULL;
			}
		}

		notein = event->byte1;
		event->byte1 = changenote(newscale,newroot,newflats,
		                          oldscale,oldroot,oldflats,event->byte1);
		if (IsNoteOn(event)) {
			tool->noteout[notein] = event->byte1;
//dprintf(" ON: @ %6ld in=%3d out=%3d\n",event->time,notein,tool->noteout[notein]);
		}
		if ((tool->moflags & BRANCH_NEW) && tool->tool.branch) {
			event->tool = tool->tool.branch;
			event->type |= EVENT_BRANCH;
		}
		else event->tool = tool->tool.next;
	}
	else if (IsSysEx(event)) {
		event = dosysex(tool,(struct StringEvent *)event);
		if (event) event->tool = tool->tool.next;
		if ((!tool->newkey) || (!tool->oldkey)) {
			functions->msgbox("Key(s) referenced in SysEx mysteriously vanished!",
			                  "Oops",NULL,NULL,1,BP_LGREY,BP_RED,BP_BLUE);
		}
	}
	return(event);
}

/*==========================================================================
* keytoindex
*
*  INPUT: pointer to Chord
* RETURN: index of input in scalelist, or -1 if not found
*
* EFFECT: scans and counts
*=========================================================================*/

static long __regargs keytoindex(struct Chord *key)

{
	register struct Chord *c;
	long i;

	if (key) {
		for ( i=1, c = functions->scalelist; c ; i++, c = c->next ) {
			if (c == key) {
//dprintf("ktoi()=%ld c=%ld\n",i,c);
				return(i);
			}
		}
	}
	return(-1); // key removed or not selected
}


/*==========================================================================
* dosnapshot
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: makes a sysex message and posts it
*=========================================================================*/

static void __regargs dosnapshot(struct KeyTool *tool)

{
	UBYTE sx[SX_LENGTH];
	struct StringEvent *event;

	*(long *)sx = keytoindex(tool->newkey);
	*(long *)(sx+4) = keytoindex(tool->oldkey);
	memcpy(sx+8,&tool->newroot,6);
	if (event = BuildSnapshot(&tool->tool, 
	                          functions->running ? functions->timenow : -1,
	                          sx, SX_LENGTH)) {
		PostSnapshot(&tool->tool,event);
		if (tool->tool.track->edit->window) {
			functions->updateobject(&functions->inittempo,0); /* redisplay */
		}
	}
}

/* for displaying key name */
#define NAME_X 84
#define NAME_LENGTH 15
/*==========================================================================
* drawkeyname
*
*  INPUT: pointer to tool, pointer to string
* RETURN: none
*
* EFFECT: write the string in the window
*=========================================================================*/

static void __regargs drawkeyname(struct KeyTool *tool, UWORD y, UBYTE active,
                                  const char *s)
{
	Move(tw->RPort,NAME_X,y);
	SetAPen(tw->RPort,active ? BP_WHITE : BP_DGREY);
	SetBPen(tw->RPort,BP_BLACK);
	SetDrMd(tw->RPort,JAM2);
	Text(tw->RPort,s,NAME_LENGTH);
}


/*==========================================================================
* edittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: ui handler
*=========================================================================*/

#define NEW_NAME_Y 23
#define OLD_NAME_Y 50

#define REFRESH_GADGETS 1
#define REFRESH_PIPES 2

static void __saveds __stdargs edittool(struct KeyTool *tool)

{
	struct IntuiMessage *message;
	long class; 
	long code;
	struct Menu *menu;
	struct MenuItem *menuitem;
	struct Gadget *gadget;
	struct NewWindow *nw;
	UBYTE refresh;
	char s[NAME_LENGTH+1];

	static char *notenames[12] = {
		"C","Db","D","Eb","E","F","F#","G","Ab","A","Bb","B"
	};
	static char defaultname[] = {
		" C Major"
	};

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);
	if (InToolPad((struct Tool *)tool)) refresh = 20;
	else refresh = 0;
	nw = OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,refresh);
	if (!nw) return;
	if (refresh) {
		MenuItem10.Flags &= ~ITEMENABLED;
		MenuItem11.Flags &= ~ITEMENABLED;
	}
	else {
		MenuItem10.Flags |= ITEMENABLED;
		MenuItem11.Flags |= ITEMENABLED;
		AddSnapshotGadget(tw);
	}
	if (menu = functions->DupeMenu(&MenuList1)) SetMenuStrip(tw,menu);
	else SetWindowTitles(tw,(UBYTE *)-1,MenuList1.MenuName);

	functions->EmbossOn(tw,NEWEVENTS,1);
	functions->EmbossOn(tw,OLDEVENTS,1);
	functions->EmbossOn(tw,NEWKEY,0);
	functions->EmbossOn(tw,OLDKEY,0);
	refresh = 0xFF;
	for ( ; ; ) {
		if (refresh & REFRESH_GADGETS) {
			functions->EnableEmbossed(tw,NEWKEY,!(tool->flags & NEW_KEY_EVENTS));
			functions->DrawEmbossed(tw,NEWKEY);
			if (findkey(tool->newkey)) {
				sprintf(s,"%2s %-12.12s",notenames[tool->newroot],
					                      functions->stringtext(tool->newkey->name));
			}
			else { /* it's gone now */
				tool->newkey = NULL;
				tool->newflats = 0;
				sprintf(s,"%-15s",defaultname);
			}
			drawkeyname(tool,NEW_NAME_Y,!(tool->flags & NEW_KEY_EVENTS),s);
			functions->EnableEmbossed(tw,OLDKEY,!(tool->flags & OLD_KEY_EVENTS));
			functions->DrawEmbossed(tw,OLDKEY);
			if (findkey(tool->oldkey)) {
				sprintf(s,"%2s %-15.15s",notenames[tool->oldroot],
					                      functions->stringtext(tool->oldkey->name));
			}
			else { /* it's gone now */
				tool->oldkey = NULL;
				tool->oldflats = 0;
				sprintf(s,"%-18s",defaultname);
			}
			drawkeyname(tool,OLD_NAME_Y,!(tool->flags & OLD_KEY_EVENTS),s);
			functions->SelectEmbossed(tw,NEWEVENTS,tool->flags & NEW_KEY_EVENTS);
			functions->DrawEmbossed(tw,NEWEVENTS);
			functions->SelectEmbossed(tw,OLDEVENTS,tool->flags & OLD_KEY_EVENTS);
			functions->DrawEmbossed(tw,OLDEVENTS);
		}
		if (refresh & REFRESH_PIPES) {
			gadget = functions->GetGadget(tw,POP_OLD);
			if (tool->moflags & BRANCH_OLD) gadget->GadgetRender = (APTR)&ElbowImage;
			else if (tool->moflags & PASS_OLD) gadget->GadgetRender = (APTR)&PipeImage;
			else gadget->GadgetRender = (APTR)&PlugImage;
			gadget = functions->GetGadget(tw,POP_NEW);
			if (tool->moflags & BRANCH_NEW) gadget->GadgetRender = (APTR)&ElbowImage;
			else gadget->GadgetRender = (APTR)&PipeImage;
			RefreshGList(gadget,tw,NULL,2);
		}
		refresh = 0;
		message = functions->GetIntuiMessage(tw);
		if (message->ExecMessage.mn_Length == SECRET_MESSAGE) {
			functions->myfree((char *)message,sizeof(struct Message));
			refresh = 0xFF;
			continue;
		}
		class = message->Class;
		code = message->Code;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW) break;

		else if (class == MENUPICK) {
			while (code != MENUNULL) {
				menuitem = ItemAddress(menu,code);
				switch (ITEMNUM(code)) {
					case 0 : // use sysex
						tool->flags ^= USE_SYSEX;
						break;
					case 1 : // free sysex
						tool->flags ^= FREE_SYSEX;
						break;
					case 2 : // snap
						dosnapshot(tool);
						break;
				}
				code = menuitem->NextSelect;
			}
		}

		else if (class == GADGETUP) {
			if (gadget->GadgetID == NEWEVENTS) {
				tool->flags ^= NEW_KEY_EVENTS;
				if (tool->flags & NEW_KEY_EVENTS) tool->flags &= ~OLD_KEY_EVENTS;
				refresh=1;
			}
			else if (gadget->GadgetID == OLDEVENTS) {
				tool->flags ^= OLD_KEY_EVENTS;
				if (tool->flags & OLD_KEY_EVENTS) tool->flags &= ~NEW_KEY_EVENTS;
				refresh=1;
			}
		}
		else if (class == GADGETDOWN) {
			if (gadget->GadgetID == NEWKEY) {
				tool->newroot = functions->popupkey(tw,tool->newroot);
				tool->newkey = functions->popupscale(tool->newkey);
				if ( (tool->newroot == 1) || (tool->newroot == 3) ||
				     (tool->newroot == 5) || (tool->newroot == 8) ||
					  (tool->newroot == 10) )
					tool->newflats = 1;
				else tool->newflats = 0;
				refresh=1;
			}
			else if (gadget->GadgetID == OLDKEY) {
				tool->oldroot = functions->popupkey(tw,tool->oldroot);
				tool->oldkey = functions->popupscale(tool->oldkey);
				if ( (tool->oldroot == 1) || (tool->oldroot == 3) ||
				     (tool->oldroot == 5) || (tool->oldroot == 8) ||
					  (tool->oldroot == 10) )
					tool->oldflats = 1;
				else tool->oldflats = 0;
				refresh=1;
			}
			else if (gadget->GadgetID == POP_NEW) {
				functions->EnableMenuItem(&PopPipeMenu,0,PLUG_ITEM,0,0); // no plug
				code = functions->PopUpMenu(tw,&PopPipeMenu);
				switch (ITEMNUM(code)) {
					case PIPE_ITEM :
						tool->moflags = (tool->moflags & ~BRANCH_NEW) | PASS_NEW;
						break;
					case ELBOW_ITEM :
						tool->moflags = (tool->moflags & ~PASS_NEW) | BRANCH_NEW;
						break;
					default : break;
				}
				functions->EnableMenuItem(&PopPipeMenu,0,1,0,1);
				refresh = REFRESH_PIPES;
			}
			else if (gadget->GadgetID == POP_OLD) {
				code = functions->PopUpMenu(tw,&PopPipeMenu);
				switch (ITEMNUM(code)) {
					case PIPE_ITEM :
						tool->moflags = (tool->moflags & ~BRANCH_OLD) | PASS_OLD;
						break;
					case PLUG_ITEM :
						tool->moflags &= ~(BRANCH_OLD | PASS_OLD);
						break;
					case ELBOW_ITEM :
						tool->moflags = (tool->moflags & ~PASS_OLD) | BRANCH_OLD;
						break;
					default : break;
				}
				refresh = REFRESH_PIPES;
			}
		}
	}

	if (menu) functions->DeleteMenu(menu);
	functions->EmbossOff(tw,NEWKEY);
	functions->EmbossOff(tw,OLDKEY);
	functions->EmbossOff(tw,NEWEVENTS);
	functions->EmbossOff(tw,OLDEVENTS);
	NewCloseEmbossedWindow(&tool->tool,nw);
}

/*==========================================================================
* loadtool
*
*  INPUT: file id, size of chunk to read
* RETURN: pointer to allocated tool, or NULL
*
* EFFECT: allocates and initializes a tool with data read from disk
*=========================================================================*/

static struct Tool * __saveds __stdargs loadtool(long file, long size)

{
	struct KeyTool *tool = (struct KeyTool *)
	                        functions->myalloc(TOOL_SIZE,MEMF_CLEAR);

	if (tool) {
		functions->fastread(file,(char *)tool,size);
		/* inittool has to find the key after a load */
		tool->tool.touched &= ~TOUCH_INIT;
		tool->flags |= FLAG_BOGUS;
		size -= (TOOL_SIZE+TABLE_SIZE);
		if ( size > 0 ) functions->fastseek(file,size,1); /* skip leftovers */
	}
	return((struct Tool *)tool);
}

/*==========================================================================
* savetool
*
*  INPUT: file id, pointer to tool to save
* RETURN: 0 if successful; 1 otherwise
*
* EFFECT: writes a tool's data to a file.
*=========================================================================*/

static long __saveds __stdargs savetool(long file, struct KeyTool *tool)

{
	struct Chord *c;
	long key_index;

	functions->fastwrite(file,(char *)(&tool->tool.toolmaster->toolid),4);
	/* then the size of the data */
	key_index = TOOL_SIZE-TABLE_SIZE;
	functions->fastwrite(file,(char *)(&key_index),4);
	functions->fastwrite(file,(char *)(&tool->tool),sizeof(struct Tool));
	key_index = keytoindex(tool->newkey);
	functions->fastwrite(file,(char *)(&key_index),4);
	key_index = keytoindex(tool->oldkey);
	functions->fastwrite(file,(char *)(&key_index),4);

	/* and the rest of the stuff less size of indices, and the lookup table */
	functions->fastwrite(file,(char *)(&tool->newroot),
	                     TOOL_SIZE-8-sizeof(struct Tool)-TABLE_SIZE);
	return(0);
}


static long __saveds __stdargs savesize(struct KeyTool *tool) 
{
	return(TOOL_SIZE-TABLE_SIZE);
}

static struct ToolMaster master;

/*==========================================================================
* inittoolmaster
*
*  INPUT: none
* RETURN: pointer to ToolMaster
*
* EFFECT: initializes.
*=========================================================================*/

struct ToolMaster * __saveds __stdargs inittoolmaster(void)

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = TOOL_SIZE;
	master.toolid = KTOR_TOOLID;
	master.edittool = edittool;
	master.processevent = processevent;
	master.processlist = processlist;
	master.savesize = savesize;
	master.loadtool = loadtool;
	master.savetool = savetool;
	master.image = &KeyolatorDown;
	master.upimage = &KeyolatorUp;
	strcpy(master.name,WTITLE);
	master.tooltype = TOOL_NORMAL | TOOL_BRANCHOUT;

	return(&master);
}
